/*
  Copyright Red Hat, Inc. 2005

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to the
  Free Software Foundation, Inc.,  675 Mass Ave, Cambridge, 
  MA 02139, USA.
*/
/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */


#ifndef Socket_h
#define Socket_h

#include "counting_auto_ptr.h"

#include "String.h"
#include <vector>


// NOT THREAD SAFE
// provide external locking


class Socket
{
 public:
  Socket(const Socket&);
  virtual Socket& operator= (const Socket&);
  virtual ~Socket();
  
  virtual bool operator== (const Socket&);
  virtual bool server() = 0;
  
  int get_sock();
  bool valid() { return _sock != -1; }
  
  bool nonblocking();  // return whether O_NONBLOCK is set
  bool nonblocking(bool mode);  // set O_NONBLOCK, return old state
  
 protected:
  Socket(int sock);  // takes ownership of sock
  int _sock;
  void close();
  counting_auto_ptr<int> _counter;
  
  void poll(bool& read, bool& write, int timeout);  // milliseconds
  
 private:
  void decrease_counter();
};  // class Socket


class ServerSocket;

class ClientSocket : public Socket
{
 public:
  ClientSocket();
  ClientSocket(const String& sock_path);  // UNIX socket
  ClientSocket(const String& hostname, unsigned short port,
	       unsigned int timeout_ms=0  /*  0 - standard blocking behavior
					     >0 - timeout
					  */ );  // TCP socket
  ClientSocket(const ClientSocket&);
  virtual ClientSocket& operator= (const ClientSocket&);
  virtual ~ClientSocket();
  
  virtual String recv();
  virtual String recv(int timeout);
  virtual String send(const String& msg);  // return what is left to send
  virtual String send(const String& msg, int timeout);
  
  virtual void ready(bool& recv, bool& send, int timeout);
  
  virtual bool server() { return false; }
  
  virtual bool connected_to(const String& hostname);
  
 protected:
  u_int32_t _addr;  // address in network byte order 
  
  ClientSocket(int sock, u_int32_t addr=0);  // takes ownership of sock
  friend class ServerSocket;
};  // ClientSocket


class ServerSocket : public Socket
{
 public:
  ServerSocket(const String& sock_path); // UNIX socket
  ServerSocket(unsigned short port); // TCP socket
  ServerSocket(const ServerSocket&);
  virtual ServerSocket& operator= (const ServerSocket&);
  virtual ~ServerSocket();
  
  ClientSocket accept();
  
  virtual bool ready(int timeout);
  
  virtual bool server() { return true; }
  
 private:
  bool _unix_sock;
  String _sock_path;
  
};  // ServerSocket


#endif  // Socket_h
