/*
  Copyright Red Hat, Inc. 2005

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to the
  Free Software Foundation, Inc.,  675 Mass Ave, Cambridge, 
  MA 02139, USA.
*/
/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */


#include "ClusterConf.h"
#include "utils.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>

#include <fstream>
using namespace std;


#define CLUSTER_CONF_DIR     String("/etc/cluster/")
#define CLUSTER_CONF_NAME    String("cluster.conf")
#define CLUSTER_CONF_PATH    (CLUSTER_CONF_DIR + CLUSTER_CONF_NAME)

#define CCS_TOOL_PATH        "/sbin/ccs_tool"


static void 
set_CMAN_TOOL_PATH();
static String CMAN_TOOL_PATH;



XMLObject
ClusterConf::get()
{
  return readXML(CLUSTER_CONF_PATH);
}

void
ClusterConf::set(const XMLObject& xml,
		 bool propagate)
{
  set_CMAN_TOOL_PATH();
  
  // sanity check
  if (xml.tag() != "cluster")
    throw String("invalid cluster.conf");
  if (xml.get_attr("name").empty())
    throw String("invalid cluster.conf");
  long long conf_version = utils::to_long(xml.get_attr("config_version"));
  if (conf_version == 0)
    throw String("invalid cluster.conf");
  
  // create dir, if not existing
  DIR* dir = opendir(CLUSTER_CONF_DIR.c_str());
  if (dir)
    closedir(dir);
  else {
    if (errno == ENOENT) {
      if (mkdir(CLUSTER_CONF_DIR.c_str(), 
		S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH))
	throw String("failed to create ") + CLUSTER_CONF_DIR;      
    } else
      throw String("opendir() error");
  }
  
  // save tmp cluster.conf
  String tmp_path(CLUSTER_CONF_PATH + ".tmp_ricci");
  fstream fs(tmp_path.c_str(), 
	     ios_base::out | ios_base::trunc);
  fs << generateXML(xml);
  fs.close();
  
  // propagate/move to proper place
  if (propagate) {
    try {
      String out, err;
      int status;
      vector<String> args;
      args.push_back("update");
      args.push_back(tmp_path);
      if (utils::execute(CCS_TOOL_PATH, args, out, err, status, false))
	throw command_not_found_error_msg(CCS_TOOL_PATH);
      if (status != 0)
	throw String("ccs_tool failed");
      
      if (is_cman(xml)) {
	args.clear();
	args.push_back("version");
	args.push_back("-r");
	args.push_back(utils::to_string(conf_version));
	if (utils::execute(CMAN_TOOL_PATH, args, out, err, status, false))
	  throw command_not_found_error_msg(CMAN_TOOL_PATH);
	if (status != 0)
	  throw String("cman_tool failed");
      }
      
      unlink(tmp_path.c_str());
    } catch ( ... ) {
      unlink(tmp_path.c_str());
      throw;
    }
  } else {
    if (rename(tmp_path.c_str(), CLUSTER_CONF_PATH.c_str())) {
      unlink(tmp_path.c_str());
      throw String("failed to rename cluster.conf");
    }
  }
}

void 
ClusterConf::purge_conf()
{
  unlink(CLUSTER_CONF_PATH.c_str());
}

bool
ClusterConf::is_gulm(const XMLObject& cluster_conf)
{
  for (list<XMLObject>::const_iterator iter = cluster_conf.children().begin();
       iter != cluster_conf.children().end();
       iter++)
    if (iter->tag() == "gulm")
      return true;
  return false;
}

bool
ClusterConf::is_cman(const XMLObject& cluster_conf)
{
  return !is_gulm(cluster_conf);
}



void 
set_CMAN_TOOL_PATH()
{
  static bool path_set = false;
  if (!path_set) {
    CMAN_TOOL_PATH = "/sbin/cman_tool";
    if (access(CMAN_TOOL_PATH.c_str(), X_OK)) 
      CMAN_TOOL_PATH = "/usr/sbin/cman_tool";
    path_set = true;
  }
}
